package be.iminds.ilabt.jfed.restapi.json;

import be.iminds.ilabt.jfed.restapi.JFedRestDerivedUriGenerator;
import be.iminds.ilabt.util.jsonld.JsonLdObjectsMetaData;
import be.iminds.ilabt.util.jsonld.UriTool;
import be.iminds.ilabt.util.jsonld.iface.JsonLdObject;
import be.iminds.ilabt.util.jsonld.impl.BasicLdObject;
import be.iminds.ilabt.util.jsonld.test.BasicObjectTestCollection;
import be.iminds.ilabt.util.jsonld.test.CommonTest;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import io.dropwizard.jackson.Jackson;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;

import java.net.URISyntaxException;
import java.util.*;

import static io.dropwizard.testing.FixtureHelpers.fixture;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.equalTo;
import static org.hamcrest.Matchers.is;

@RunWith(Parameterized.class)
public class JsonTest {
    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();

    private static JsonLdObjectsMetaData instance = null;
    private static JsonLdObjectsMetaData getJsonLdObjectsMetaData() {
        if (instance == null) {
            instance = JsonLdObjectsMetaData.getInstance(Experiment.class.getPackage().getName());
        }
        return instance;
    }

    @Parameterized.Parameters(name = "{index}: collection={0} index={1}")
    public static Collection<Object[]> data() throws Exception {
        List<Object[]> res = new ArrayList<>();

        ExperimentTestCollection experimentTestCollection = new ExperimentTestCollection();
        for (int i = 0; i < experimentTestCollection.getSize(); i++) {
            res.add(new Object[]{experimentTestCollection, i});
        }

        SpeaksForCredentialTestCollection speaksForCredentialTestCollection = new SpeaksForCredentialTestCollection();
        for (int i = 0; i < speaksForCredentialTestCollection.getSize(); i++) {
            res.add(new Object[]{speaksForCredentialTestCollection, i});
        }

        return res;
    }

    private final BasicObjectTestCollection testCollection;
    private final int index;
    public JsonTest(BasicObjectTestCollection testCollection, int index) throws URISyntaxException {
        UriTool uriTool = getJsonLdObjectsMetaData().makeUriTool("http://localhost:9000/api", new JFedRestDerivedUriGenerator());
        if (index == 0) {
            (testCollection).setUri(0, uriTool);
        }
        this.testCollection = testCollection;
        this.index = index;
    }

    @Test
    public void serializesToJSON() throws Exception {
        String fixtureFilename = testCollection.getFixtureByIndex(index);
        System.out.println("serializesToJSON is reading " + fixtureFilename);
        JsonLdObject obj = testCollection.getByIndex(index);

        final String expected =  MAPPER.writer().with(SerializationFeature.INDENT_OUTPUT).writeValueAsString(
                MAPPER.readValue(
                        fixture(fixtureFilename),
                        obj.getClass())
        );

        final String actual = MAPPER.writer().with(SerializationFeature.INDENT_OUTPUT).writeValueAsString(obj);
        assertThat(actual, is(equalTo(expected)));
    }

    @Test
    public void serializesToJSONTree() throws Exception {
        String fixtureFilename = testCollection.getFixtureByIndex(index);
        System.out.println("serializesToJSON is reading " + fixtureFilename);
        JsonLdObject exp = testCollection.getByIndex(index);

        final Map expected = MAPPER.readValue(fixture(fixtureFilename), Map.class);
        final Map actual = MAPPER.readValue(MAPPER.writer().with(SerializationFeature.INDENT_OUTPUT).writeValueAsString(exp), Map.class);

        assertThat(mapToString(actual), is(equalTo(mapToString(expected))));
        assertThat(mapToString(actual) + "\n\n != \n\n" + mapToString(expected), actual, is(equalTo(expected)));
    }

    @Test
    public void deserializesFromJSON() throws Exception {
        String fixtureFilename = testCollection.getFixtureByIndex(index);
        System.out.println("deserializesFromJSON is reading " + fixtureFilename);
        JsonLdObject exp = testCollection.getByIndex(index);
        JsonLdObject actualFedmonObject = MAPPER.readValue(fixture(fixtureFilename), exp.getClass());

//        assertThat(actualFedmonObject, is(equalTo(expectedFedmonObject)));

//        System.out.flush();
//        System.err.flush();
//        System.out.println("deserializesFromJSON is comparing actual="+actualFedmonObject+" expected="+expectedFedmonObject);
//        System.out.flush();
//        System.err.flush();
        testCollection.assertSame(
                actualFedmonObject, exp,
                index == 0 ? CommonTest.ComparePrecision.FULL : CommonTest.ComparePrecision.IGNORE_URI,
                index == 0 ? CommonTest.ComparePrecision.FULL : CommonTest.ComparePrecision.IGNORE_COMPLETELY);
        System.out.flush();
        System.err.flush();
    }

    /** helper for nice map comparison */
    private static String mapToString(Map m) {
        if (m == null) return "null";
        String res = "{\n";
        TreeSet sortedKeys = new TreeSet(m.keySet());
        boolean first = true;
        for (Object key : sortedKeys) {
            if (!first)
                res += ",\n";
            first = false;
            Object val = m.get(key);
            if (val== null) {
                res += "  \"" + key + "\" : null";
                continue;
            }
            if (val instanceof Number || val instanceof Boolean) {
                res += "  \"" + key + "\" : " + val;
                continue;
            }
            if (val instanceof String) {
                res += "  \"" + key + "\" : \"" + val+"\"";
                continue;
            }
            if (val instanceof Map) {
                res += "  \"" + key + "\" : " + mapToString((Map) val);
                continue;
            }
            if (val instanceof List) {
                res += "  \"" + key + "\" : " + listToString((List) val);
                continue;
            }
            throw new RuntimeException("Unsupported value type: "+val.getClass().getName());
        }
        return res+"\n}";
    }
    private static String listToString(List l) {
        if (l == null) return "null";
        String res = "[\n";
        boolean first = true;
        for (Object val : l) {
            if (!first)
                res += ", ";
            first = false;
            if (val== null) {
                res += "null";
                continue;
            }
            if (val instanceof Number || val instanceof Boolean) {
                res += val;
                continue;
            }
            if (val instanceof String) {
                res += "\"" + val+"\"";
                continue;
            }
            if (val instanceof Map) {
                res +=  mapToString((Map) val);
                continue;
            }
            if (val instanceof List) {
                res += listToString((List)val);
                continue;
            }
            throw new RuntimeException("Unsupported value type: "+val.getClass().getName());
        }
        return res + "\n]";

    }
}
